//
// SHDesignes CoExec library version 1.4
//
// Copyright 2006-2008 SHDesigns
//
// This code can be used freely and with no cost. The source can not
// be sold without permission from SHDesigns.
//
// please consider donating to support this project
// donations can be made via paypal to rabbit@shdesigns.org
//
#include "..\co_exec.h"
#include <string.h>
#include <rabbit.h>
//#define SAVE_ALL

TASK * _cur_task=tasks;
TASK * _last_task=tasks;
TASK * _priority_task=NULL;

static unsigned char max_tasks=0;
unsigned char task_count=0;  // first task is always present

#ifdef FAR_STACKS
static char far * fstacks=(char far *)0l;
static long stack_bsize=0l;

int init_tasks(int num_t,long fstacksize)
#else
void init_tasks(int num_t)
#endif
{
    char * stack,*stack_end;
    max_tasks=num_t;
    memset(tasks,0,sizeof(TASK)*num_t);
    tasks[0].run=1; // first yeild will create the stack pointer
#ifdef FAR_STACKS
	fstacks=(char far *)xalloc(fstacksize);
	if (fstacks==(char far *)0l)
		return -1;
    stack_bsize=fstacksize;
#asm
	ioi	ld	a,(STACKSEG)
	ioi ld (DATASEG),a 	; remap dataseg to be same as stackseg
	ioi	ld	a,(SEGSIZE)
	or		a        ; clr carry
	rra
	rra
	rra
	rra
	or		0xd0
	ioi ld	(SEGSIZE),a		;
#endasm
#endif
#asm
    bool    hl
    ld      l,h
    add     hl,sp
    push    hl
    pop     bc
    ld      hl,_stackMarker##
    ld      de,4
    add     hl,de
#endasm
    stack_end=(char *)_BC;
    stack=(char *)_HL; // skip over stack marker
    stack_end-=10;
    memset(stack,0x00,stack_end-stack);
    // fake out start values, the reasl stack entry will be
    // entered in on first task switch
    // this allows sane values if main task gets restarted.
    tasks[0].stack_start=(unsigned *)stack;
    tasks[0].stack_size=stack_end-stack;
    tasks[0].stack=(unsigned *)stack_end;
	_cur_task=_last_task=tasks;
#ifdef FAR_STACKS
	tasks[0].sseg=ini(STACKSEG);
	return 0;
#endif
}

//static unsigned * _SP;
static char _in_scheduler=1; // set to 1 causes no switching until ready
#pragma offset_labels(TASK)

void _nearcall _nostep task_switch(void)
{
#asm
    push    ip
	ld		hl,__in_scheduler
	inc		(hl)
	ld   	a,(hl)
	dec		a
	jp		nz,.exit_ts_busy
    push    iy
    push    ix
	ld		h,a
	ld		l,a
    add		hl,sp
	ld		ix,(__cur_task)
	ld		(ix+STACK_OFF),hl
	ld		hl,(__priority_task)
	ld		a,h
	or		l
	jr		z,.next_task
	ld		ix,hl		; priority is next
	bool	hl
	ld		l,h
	ld		(__priority_task),hl
.next_task:
	ld		hl,ix
	ld		bc,0+._task.exit+2
	add		hl,bc
	ex		de,hl
	ld		hl,(__last_task)
	xor	a
	sbc		hl,de
	jr		nc,.no_wrap
	ld		de,_tasks
.no_wrap:
	ld		(__cur_task),de	// used in next line
#if 0
#endasm
    while (!_cur_task->run)
	{
        if ((_cur_task->delay)&&(getMilliSeconds()-_cur_task->delay_start)>=_cur_task->delay)
            _cur_task->run=1;
		else
			if (++_cur_task>_last_task)
	   			_cur_task=tasks;
	}
	;
#asm
#endif
   ex		de,hl
   ld	  	ix,hl
search_tasks:
   bit	0,(ix+RUN_OFF)
   jr		nz,run_it
   ld		hl,(ix+DELAY_OFF)
   bool	hl
   jr		nz,.do_delay
   ld		hl,(ix+DELAY_OFF+2)
   bool	hl
   jr		z,.next_task
.do_delay:
;	_tdiff=MS_TIMER-_cur_task->delay_start;
   ld		de,(__TimerMilliSeconds+4)
   ld    hl,(ix+DELAY_START_OFF)
   ex		de,hl
   xor	a
   sbc	hl,de
   ex		de',hl
   ld		de,(__TimerMilliSeconds+6)
   ld    hl,(ix+DELAY_START_OFF+2)
   ex		de,hl
   sbc	hl,de
   ex		de,hl
   ld		bc',de
;	_tdiff-=_cur_task->delay;
	exx        	; BC:DE is difference
   ld    hl,(ix+DELAY_OFF)
   ex		de,hl
   xor	a
   sbc	hl,de
   ld   	d,b
   ld		e,c
   ld    hl,(ix+DELAY_OFF+2)
   ex		de,hl
   sbc	hl,de
   jp		c,.next_task
   bool	hl
   ld		l,h
   ld		(ix+DELAY_OFF),hl
   ld		(ix+DELAY_OFF+2),hl
	ld		(ix+RUN_OFF),1
run_it:
;	ld		ix,(_cur_task)
	ld	ix,(__cur_task)
	ld	hl,(ix+STACK_OFF)
#ifdef FAR_STACKS
	ld	a,(ix+SSEG_OFF)
#endif
    ld      sp,HL	; this will disable ints for the next instruction
#ifdef FAR_STACKS
	ioi ld (STACKSEG),a
#endif
	ipres
    pop     ix
    pop     iy
	ld		hl,__in_scheduler
.exit_ts_busy:
	dec		(hl)
    pop     ip
	nop			; just for debugger
#endasm
}

void sleep(unsigned long stime)
{
    if (stime==0l)
    {   // 0 time just do a task switch
        yield;
        return;
    }
	// this is if the exec is not running, use the delay
    if (task_count==0)
    {
        while (stime>30000)
        {
            delay(30000);
            stime-=30000;
        }
        delay((int)stime);
        return;
    }
    _cur_task->delay=stime;
    _cur_task->delay_start=getMilliSeconds();
    _cur_task->run=0;    // say not to run
    yield;  // will wait delay before return;
}

//
// this will cause a near return to the task function
//
#pragma offset_labels on
#pragma warn -759
void near task_start(int (*task)(),int param) // task is pointer to function
{
    int retval;
    _DE=param;
    _HL=(unsigned)task;
#asm
    push    ix
    push    de
    ld      de,fexit
    push    de
    jp      (hl)
fexit:
    pop     de  ; get param
    pop     ix
    ld  (ix+.retval),hl
#endasm
    _cur_task->run=0;
    _cur_task->exit=retval;
#ifdef DEBUG
    printf("Task 0x%x exited retval=%d!\r\n",task,retval);
#endif
    task_switch();  // should never return;
}
#pragma warn 759

static int x_create_task(int tnum,int (* func)(int),void * stack,unsigned stack_size,int param)
{
#ifdef FAR_STACKS
#define SSFAR far
	unsigned long scalc;
	unsigned fsstart,sseg;
	TASK_STACK far * tstack;
    unsigned char far * new_stack;
	if (stack_size==0) // this is a restart
	{
		scalc=(unsigned long)tasks[tnum].stack_start;
    	scalc+=((unsigned long)tasks[tnum].sseg)<<12;
    	stack_size=tasks[tnum].stack_size;
		new_stack=(unsigned char far *)scalc;
		sseg=(unsigned)(((unsigned long)new_stack-0xd000l)>>12);
		fsstart=0xd000|(((unsigned)new_stack)&0x0fff);
	}
	else
	{
		if (stack!=NULL) // near stack
		{
			new_stack=stack;
			sseg=ini(STACKSEG);
			fsstart=(unsigned)stack;
		}
		else
		{
			new_stack=(unsigned char far*)fstacks;
			stack_bsize-=stack_size;
			if (stack_bsize<0)
			{
				stack_bsize+=stack_size;
				return -1;
			}
			if (((unsigned long)new_stack&0xff000l)!=(((unsigned long)new_stack+stack_size-1l)&0xff000l))
			 {
	         	// stack crosses 4k boundary allocate dummy to align
				stack_bsize-=4096l-((unsigned long)new_stack&0xfffl);
				if (stack_bsize<=0)
				{
					stack_bsize+=4096l-((unsigned long)new_stack&0xfffl);
					stack_bsize+=stack_size;
					return -1;
				}
			    fstacks+=4096l-((unsigned long)new_stack&0xfffl); // wasted mem
			    new_stack=(unsigned char far*)fstacks;
				fstacks+=stack_size;
			}
			else
				fstacks+=stack_size;
			sseg=(unsigned)(((unsigned long)new_stack-0xd000l)>>12);
			fsstart=0xd000|(((unsigned)new_stack)&0x0fff);
		}
	}
#else
#define SSFAR
	TASK_STACK * tstack;
	 unsigned char * new_stack=(unsigned char *) stack;
	if (stack==NULL)
	{
	    stack=(void *)new_stack=tasks[0].stack_start;
	    tasks[0].stack_start+=stack_size/2;
	    tasks[0].stack_size-=stack_size;
	}
#endif
    if (tnum>=(int)max_tasks)
        return -1;
    stack_size&=0xfffe; // make # words
    if (new_stack==NULL)
        return -2;
    fmemset(new_stack,0x00,stack_size);
    new_stack+=stack_size-sizeof(TASK_STACK);    // move to end of stack
#ifdef FAR_STACKS
	tstack=(TASK_STACK far *) new_stack;
#else
	tstack=(TASK_STACK *) new_stack;
#endif
	tstack->task_num=tnum;
	tstack->param=param;
	tstack->func=func;
	tstack->stub=task_start;
    memset(&tasks[tnum],0,sizeof(TASK));
	//
	// init task struct. Most values are needed for restart
	//
    tasks[tnum].exit=0;
    tasks[tnum].task=func;
    tasks[tnum].stack_size=stack_size;
#ifdef FAR_STACKS
	tasks[tnum].sseg=(char)sseg;
    tasks[tnum].stack_start=(unsigned *)fsstart;
	scalc=(unsigned long)new_stack;
	if (stack!=NULL) // near stack
	{
        scalc-=((unsigned long)ini(STACKSEG))<<12;
	}
	else
	{
		scalc=(scalc&0x0fffl)|0xd000l;
	}
    tasks[tnum].stack=(unsigned *)scalc;
#else
    tasks[tnum].stack=(unsigned *)(new_stack);
    tasks[tnum].stack_start=(unsigned *)stack;
#endif
#ifdef DEBUG
#ifdef FAR_STACKS
    printf("Create task #%d at 0x%x, %d byte stack at %02X:%04X,param=%d\r\n",tnum,func,
		stack_size<<1,sseg,tasks[tnum].stack_start,param);
#else
    printf("Create task #%d at 0x%x, %d byte stack at %x,param=%d\r\n",tnum,
		func,stack_size<<1,stack,param);
#endif
#endif
    tasks[tnum].run=1;
    return tnum;
}

void task_run(int tasknum,int run)
{
#ifdef DEBUG
    printf("Task %d run set to %d\r\n",tasknum,run);
#endif
    tasks[tasknum].run=run;
    tasks[tasknum].delay=0l; // prevent delay from restarting task
}

int restart_task(int tasknum,int param)
{
    int retval;
	task_run(tasknum,0);	// prevent it from running
#ifdef DEBUG
    printf("Task %d restarted\r\n",tasknum);
#endif
#ifdef FAR_STACKS
    retval=x_create_task(tasknum,tasks[tasknum].task,NULL,0,param);
#else
    retval=x_create_task(tasknum,tasks[tasknum].task,tasks[tasknum].stack_start,tasks[tasknum].stack_size,param);
#endif
    return retval;
}

int create_task(int (* func)(int),void * stack,unsigned stack_size,int param)
{
	int tnum;
    if (task_count==0)
    {
	    task_count=1;   // first task is main loop.
	    _in_scheduler=0;
	}
	tnum=x_create_task(task_count,func,stack,stack_size,param);
	if (tnum==(int)task_count)
	{
		_last_task=tasks+task_count;
		task_count++;
	}
	return tnum;
}

void task_abort(int tasknum)
{
#ifdef DEBUG
    printf("Task %d set to abort\r\n",tasknum);
#endif
    if (tasknum<0)
    {
        for (unsigned char i=1;i<task_count;i++)
          tasks[i].abort=1; // tell all tasks to exit
    }
    else
        tasks[tasknum].abort=1;
}

int near check_abort(void)
{
    return _cur_task->abort;
}

int check_exit(int task_num)
{
    if (tasks[task_num].run)
        return TASK_RUNNING;
    return tasks[task_num].exit;
}



